# Brining some "prompt generator" classes - note that you can derive and extend those
from promptimize.prompt_cases import PromptCase

# Bringing some useful eval function that help evaluating and scoring responses
# eval functions have a handle on the prompt object and are expected
# to return a score between 0 and 1
from promptimize.utils import extract_json_objects
from preql_nlp.constants import logger


def run_prompt(prompt: str, debug: bool = False) -> list[dict | list]:
    # we probably don't need to use promptimize here
    # but consistency makes it easy to optimize with the test cases
    if debug:
        logger.debug("prompt")
        logger.debug(prompt)
    deep_prompt = PromptCase(user_input=prompt, evaluators=None)

    raw_output = deep_prompt._run(dry_run=False)
    if debug:
        logger.debug("output")
        logger.debug(raw_output)

    base = extract_json_objects(raw_output)
    if debug:
        logger.debug(base)
    return base



from preql.core.models import (
    Select,
    ProcessedQuery,
    Concept,
    OrderItem,
    OrderBy,
    Ordering,
)
from preql.core.query_processor import process_query_v2
from preql.core.enums import Purpose
from typing import Iterable, List
from collections import defaultdict

from preql.core.models import Environment
from preql_nlp.prompts import (
    run_prompt,
    gen_extraction_prompt_v1,
    gen_structured_prompt_v1,
    gen_selection_v1,
)
from preql_nlp.constants import logger, DEFAULT_LIMIT

from dataclasses import dataclass
from typing import Any

import re


def split_to_tokens(input_text: str) -> list[str]:
    return list(set(re.split("\.|\_", input_text)))


def build_token_list_by_purpose(concepts, purposes: Iterable[Purpose]):
    concepts = {k: v for k, v in concepts.items() if v.purpose in purposes}

    unique = set(concepts.keys())
    final = set()
    # if a concept has another concept as a substring, remove
    # to help restrict the size of the search space
    for concept in unique:
        # matched = [comparison  for comparison in unique if comparison in concept]
        # filtered = [m for m in matched if m !=concept]
        # if not filtered:
        #     final.append(concept)
        for x in split_to_tokens(concept):
            final.add(x)
    return ", ".join(list(final))


def tokens_to_concept(
    tokens: list[str],
    concepts: List[str],
    limits: int = 5,
    universe: list[str] | None = None,
):
    tokens = list(set(tokens))
    universe_list = list(set(universe or []))
    mappings = {x: split_to_tokens(x) for x in concepts}
    candidates = [x for x in concepts if all(token in mappings[x] for token in tokens)]
    pickings = defaultdict(list)

    if not candidates:
        return None
    tiers = set()
    for candidate in candidates:
        tier = sum(
            [1 if token in mappings[candidate] else 0 for token in universe_list]
        )
        pickings[tier].append(candidate)
        tiers.add(tier)
    tier_list = sorted(list(tiers), key=lambda x: -x)

    found: List[str] = []
    while len(found) < limits and tier_list:
        stier = tier_list.pop(0)
        candidates = sorted(pickings[stier], key=lambda x: len(x))
        required = limits - len(found)
        found += candidates[:required]
    return found


@dataclass
class IntermediateParseResults:
    select: List[str]
    limit: int
    order: List[str]
    # filter: List[FilterItem]


def coerce_list_dict(input: Any) -> List[dict]:
    if not isinstance(input, list):
        raise ValueError("Input must be a list")
    for x in input:
        if not isinstance(x, dict):
            raise ValueError("Input must be a list of dicts")
    return input


def coerce_list_str(input: Any) -> List[str]:
    if not isinstance(input, list):
        raise ValueError("Input must be a list")
    for x in input:
        if not isinstance(x, str):
            raise ValueError("Input must be a list of dicts")
    return input


def discover_inputs(
    input_text: str, input_environment: Environment, debug: bool = False
) -> IntermediateParseResults:
    # we work around promopt size issues and hallucination by doing a two phase discovery
    # first we parse the question into metrics/dimensions
    # then for each one, we match those to a token list
    # and then we deterministically map the tokens back to the most relevant concept
    # TODO: turn the third pass into a prompt
    concepts = input_environment.concepts
    debug = True
    final_concept_list = list(concepts.keys())
    metrics = build_token_list_by_purpose(concepts, (Purpose.METRIC,))
    dimensions = build_token_list_by_purpose(
        concepts, (Purpose.KEY, Purpose.PROPERTY, Purpose.CONSTANT)
    )

    parsed = coerce_list_dict(
        run_prompt(gen_extraction_prompt_v1(input_text), debug=debug)
    )[0]
    order = parsed.get("order", [])
    token_inputs = {"metrics": metrics, "dimensions": dimensions}

    output: List[str] = []
    for field in [
        "metrics",
        "dimensions",
    ]:
        local_phrases = [x for x in parsed[field]]
        phrase_tokens = coerce_list_dict(
            run_prompt(
                gen_structured_prompt_v1(
                    phrases=local_phrases, tokens=token_inputs[field]
                ),
                debug=True,
            )
        )
        token_universe = []
        for mapping in phrase_tokens:
            for k, v in mapping.items():
                token_universe += v
        for mapping in phrase_tokens:
            for k, v in mapping.items():
                concepts = tokens_to_concept(
                    v,
                    [c for c in final_concept_list],
                    limits=5,
                    universe=token_universe,
                )
                logger.info(f"For phrase {k} got {concepts}")
                if concepts:
                    output += concepts
                else:
                    raise ValueError(
                        f"Could not find concept for input {k} with tokens {v}"
                    )
    selections = coerce_list_dict(
        run_prompt(gen_selection_v1(concepts=output, question=input_text), debug=debug)
    )[0]
    final = list(set(selections.get("matches", [])))

    return IntermediateParseResults(
        select=final, limit=parsed.get("limit", 20), order=order
    )


def safe_limit(input: int | None) -> int:
    if not input:
        return DEFAULT_LIMIT
    if input in (-1, 0):
        return DEFAULT_LIMIT
    return input


def parse_order(input: List[Concept], ordering: List[str] | None) -> OrderBy:
    default_order = [
        OrderItem(expr=c, order=Ordering.DESCENDING)
        for c in input
        if c.purpose == Purpose.METRIC
    ]
    if not ordering:
        return OrderBy(default_order)
    final = []
    for order in ordering:
        concept, direction = order.split(" ", 1)
        matched = [c for c in input if concept in c.address]
        if not matched:
            continue
        final.append(OrderItem(expr=matched[0], order=Ordering(direction)))
    return OrderBy(items=final)


def parse_query(
    input_text: str,
    input_environment: Environment,
    debug: bool = False,
):
    results = discover_inputs(input_text, input_environment, debug=debug)
    concepts = [input_environment.concepts[x] for x in results.select]
    order = parse_order(concepts, results.order)
    if debug:
        print("Concepts found")
        for c in concepts:
            print(c.address)
    query = Select(selection=concepts, limit=safe_limit(results.limit), order_by=order)
    return query


def build_query(
    input_text: str,
    input_environment: Environment,
    debug: bool = False,
) -> ProcessedQuery:
    query = parse_query(input_text, input_environment, debug=debug)
    return process_query_v2(statement=query, environment=input_environment)



from typing import List


def gen_structured_prompt_v1(phrases:List[str], tokens:str)->str:
    phrase_str = ', '.join(phrases)
    user_input = f"Given the tokens {tokens}, match tokens to the phrases {phrase_str}"
    template = f"""\
System: you are an AI that helps a user map phrases to tokens by associating a phrase with tokens related to the words
in the phrase.
Guidelines:
* you can assume the user will always provide a list of phrases
* you can assume the user will always provide a list of tokens
* only return tokens provided by the user
The output should be a VALID JSON blob with the phrases as keys and arrays of tokens as values:
* If a phrase has no matches, return an empty array
User: given the tokens ["color", "product", "year", "revenue"], match tokens to the phrases ["product revenue", "product color", "product revenue by year", "yearly revenue"]
System:
{{
    "product revenue": ["product", "revenue"],
    "product color": ["product", "color"],
    "product revenue by year": ["product", "revenue", "year"],
    "yearly revenue": ["year", "revenue"]
}}
User: given the tokens ["product", "count", "order", "year"], match tokens to the phrases ["products sold", "orders"]
System:
{{
    "products sold": ["product", "count"],
    "orders": ["order", "count"]
}}
User: { user_input }
System:
"""  # noqa"""
    return template




from .prompt_executor import run_prompt
from .query_semantic_extraction import gen_extraction_prompt_v1
from .semantic_to_tokens import gen_structured_prompt_v1
from .final_selection import gen_selection_v1
__all__ = ["run_prompt", "gen_extraction_prompt_v1", "gen_structured_prompt_v1", "gen_selection_v1"]


def gen_extraction_prompt_v1(input: str):
    return f"""
System: You are a helpful AI that translates ambiguous business questions into structured outputs.
For a provided question, you will determine if there are metrics or aggregates or dimensions,
as well as any limit, order, of filtering. 

The output should be a VALID JSON blob with the following keys and values:
- metrics: a list of concepts from the question that should be aggregated
- dimensions: a list of concepts from the question which are not metrics
- limit: a number of records to limit the results to, -1 if none specified
- order: a list of fields to order the results by, with the option to specify ascending or descending
- filtering: a list of criteria to restrict the results by


User: "What are the top 10 products by sales?"
System:
{{
"metrics":["sales"],
"dimensions": ["products"],
"limit": 10,
"order": ["sales desc"],
"filtering": []
}}

User: "What are the sales by line of business and state?"
System:
{{
"metrics":["average sales"],
"dimensions": ["line of business", "state],
"limit": -1,
"order": [],
"filtering": []
}}

User: "What is the average sales by state?"
System:
{{
"metrics":["average state sales"],
"dimensions": [],
"limit": -1,
"order": [],
"filtering": []
}}

User: "{input}"
System:

"""



# type: ignore
import ast
import re

import setuptools

_version_re = re.compile(r"__version__\s+=\s+(.*)")
with open("preql_nlp/__init__.py", "rb") as f:
    _match = _version_re.search(f.read().decode("utf-8"))
    if _match is None:
        print("No version found")
        raise SystemExit(1)
    version = str(ast.literal_eval(_match.group(1)))


with open("requirements.txt", "r") as f:
    install_requires = [line.strip().replace("==", ">=") for line in f.readlines()]

setuptools.setup(
    name="pypreql-nlp",
    version=version,
    url="",
    author="",
    author_email="pypreql-community@gmail.com",
    description="NLP interface for pypreql.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(
        exclude=[
            "dist",
            "build",
            "*.tests",
            "*.tests.*",
            "tests.*",
            "tests",
            "docs",
            ".github",
            "",
            "examples",
        ]
    ),
    package_data={
        "": ["*.tf", "*.jinja", "py.typed"],
    },
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



from promptimize.prompt_cases import BasePromptCase, utils

# patched method while waiting for upstream PR to be merged

def patch_promptimize():
    def test(self):
        test_results = []
        for evaluator in self.evaluators:
            result = evaluator(self.response)
            if not (utils.is_numeric(result) and 0 <= result <= 1):
                raise Exception("Value should be between 0 and 1")
            test_results.append(result)

        if len(test_results):
            self.execution.score = sum(test_results) / len(test_results)
            self.execution.results = test_results
        self.was_tested = True


    BasePromptCase.test = test


def process_concept(concept: str):
    return concept


def gen_selection_v1(concepts: list[str], question: str):
    concept_string = ", ".join(
        [f'"{process_concept(concept)}"' for concept in concepts]
    )
    return f"""
System: You are a helpful AI that selects the most relevant matching concepts to answer a question from a provided list.

Guidelines:
* you can assume the user will always provide a list of phrases
* you can assume the user will always provide a question
* only return concepts provided by the user
* concepts are dot seperated and in quotes, e.g. "sales" or "product.revenue"
* return the concepts that together best match the user question
* reason about each match step by step, e.g. "first match the concept 'product' to the word 'product' in the question, then match the concept 'revenue' to the word 'revenue' in the question, and together these enable aggregating revenue by year"
The output should be a VALID JSON blob with the following keys and values:
* matches: a list of concepts from the user provided list that together best match the 
* reasoning: a string explaining your step by step reasoning for the matches
User: concepts: ["product.color", "order.state", "order.year", "order.revenue.sum", "order.revenue.avg", "product.name", "order.month", "order.day", "product.manufacturer"] question: "what product colors had the most revenue in 2024?"]
System:
{{"matches": ["product.color", "order.revenue.sum", "order.year" ],
"reasoning": "product.color matches the user request for product colors, and order revenue sum would enable aggregating revenue to the color. Order year is required to filter to 2024." }}
User: concepts: ["product.color", "order.state", "order.year", "order.revenue.sum", "order.revenue.avg", "product.name", "order.month", "order.day", "product.manufacturer"] question: "What are the sales by state?"
System:
{{"matches": ["order.state", "order.revenue.sum"],
"reasoning": "order.state is the best match for state when looking at revenue, and order.revenue.sum would enable aggregating revenue." }}
User: concepts: ["product.color", "order.state", "order.year", "order.revenue.sum", "order.revenue.avg", "product.name", "order.month", "order.day", "product.manufacturer"]  question: "What are the average sales by state?"
System:
{{"matches": ["order.state", "order.revenue.avg"],
"reasoning": "order.state is the best match for state, and order.revenue.avg would capture average revenue." }}
User: concepts:"[{concept_string}]" question: "{question}"
System:

"""
